#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <bitset.h>

/* This benchmarks the bitset_count routine.  */

static double
random_uniform (double range, double mean)
{
    /* This returns a white uniformly distributed random number
       with a specified range and mean.  */

    return (drand48 () - 0.5) * range + mean;
}


static double 
random_gaussian (double variance, double mean)
{
    /* This returns a white Gaussian distributed random number with a
       specified variance and mean.  */

    double uniform1;
    double uniform2;
    
    uniform1 = drand48 ();  
    uniform2 = drand48 ();  
    return sqrt (-2.0 * variance * log (uniform1)) 
      * cos (2.0 * M_PI * uniform2) + mean;
}


static void
choose (unsigned int *seq, unsigned int bits, unsigned int pop,
	unsigned int rlsd)
{
  unsigned int i;
  bitset set;

  set = bitset_alloc (pop, BITSET_ARRAY);
  bitset_ones (set);

  i = 0;
  while (!bitset_empty_p (set))
    {
      int pos;
      int len;
      int start;
      int stop;
      int j;
      
      pos = random_uniform (bits, bits / 2);

      /* Choose run length.  */
      len = abs (random_gaussian (rlsd * rlsd, 0));

      start = pos - len / 2;
      stop = start + len - 1;
      if (start < 0)
	start = 0;
      if (stop >= bits)
	stop = bits - 1;

      len = stop - start + 1;
      if (len > (bits - i))
	len = bits - i;

      for (j = 0; j < len; j++)
	{
	  if (bitset_test (set, start + j))
	    {
	      bitset_reset (set, start + j);
	      seq[i++] = start + j;
	    }
	}
    }
}


static void
fill (bitset set, unsigned int pop, unsigned int *seq)
{
  unsigned int i;

  for (i = 0; i < pop; i++)
    bitset_set (set, seq[i]);
}


static double
run (enum bitset_type type, unsigned int bits, unsigned int pop, 
     unsigned int rlsd, unsigned int *seq)
{
  bitset set1;
  clock_t start;
  clock_t stop;
  unsigned int i;

  set1 = bitset_alloc (bits, type);

  choose (seq, bits, pop, rlsd);
  fill (set1, pop, seq);

  start = clock ();
  for (i = 0; (clock () - start) < CLOCKS_PER_SEC / 10; i++)
      if (bitset_count (set1) != pop)
	  abort ();
  stop = clock ();

  bitset_free (set1);

  return (stop - start) / (double) CLOCKS_PER_SEC * 1e6 / i;
}


static void
test (enum bitset_type type, unsigned int bits, unsigned int pop, 
      unsigned int rlsd, unsigned int *seq)
{
  const char *name;
  double runtime;
  double runtimesum;
  double runtimesumsq;
  double runtimemean;
  double runtimesd;
  bitset set1;
  int i;
  int trials = 20;

  set1 = bitset_alloc (bits, type);
  name = bitset_type_name_get (set1);
  bitset_free (set1);

  runtimesum = 0;
  runtimesumsq = 0;

  for (i = 0; i < trials; i++)
  {
      runtime = run (type, bits, pop, rlsd, seq);
      runtimesum += runtime;
      runtimesumsq += runtime * runtime;
  }

  runtimemean = runtimesum / trials;
  runtimesd = sqrt ((runtimesumsq - trials * runtimemean * runtimemean) / (trials - 1));

  printf ("%s: time mean = %.1f us, time sd = %.1f us,"
	  " bits = %d, pop = %d, sd = %d\n",
	  name, runtimemean, runtimesd, bits, pop, rlsd);
}


int
main (int argc, char **argv)
{
  unsigned int bits;
  unsigned int pop;
  unsigned int seed;
  unsigned int rlsd;
  unsigned int *seq;
  int do_all = 1;

  /* Size of bitset.  */
  bits = argc > 1 ? atoi (argv[1]) : 1000000;

  /* Population (number of seq set) in bitset.  */
  pop = argc > 2 ? atoi (argv[2]) : bits;
  if (pop > bits)
      pop = bits;

  /* Standard deviation of bit clustering. */
  rlsd = argc > 3 ? atoi (argv[3]) : 100;

  /* Pseudorandom number seed.  */
  seed = argc > 4 ? atoi (argv[4]) : 0;

  srand48 (seed);
  
  // bitset_stats_enable ();

  seq = calloc (bits, sizeof (*seq));
  choose (seq, bits, pop, rlsd);

  if (argc > 5)
    {
      do_all = 0;

      switch (argv[5][0])
	{
	case 'a':
	  test (BITSET_ARRAY, bits, pop, rlsd, seq);
	  break;
	case 'l':
	  test (BITSET_LIST, bits, pop, rlsd, seq);
	  break;
	case 'e':
	case 't':
	  test (BITSET_TABLE, bits, pop, rlsd, seq);
	  break;
	case 'v':
	  test (BITSET_VARRAY, bits, pop, rlsd, seq);
	  break;

	default:
	  do_all = 1;
	  break;
	}
    }

  if (do_all)
    {
      test (BITSET_ARRAY, bits, pop, rlsd, seq);
      test (BITSET_LIST, bits, pop, rlsd, seq);
      test (BITSET_TABLE, bits, pop, rlsd, seq);
      test (BITSET_VARRAY, bits, pop, rlsd, seq);
    }

  free (seq);

  bitset_stats_dump (stderr);
  return 0;
}
