#include <stdlib.h>
#include <bitset.h>

static void
test_set (bitset set1, bitset set2)
{
  bitset_zero (set1);
  bitset_zero (set2);

  bitset_set (set1, 127);
  if (bitset_count (set1) != 1)
    abort ();
  bitset_set (set1, 511);
  if (bitset_count (set1) != 2)
    abort ();
  bitset_set (set1, 4095);
  if (bitset_count (set1) != 3)
    abort ();

  if (!bitset_test (set1, 127))
    abort ();
  if (!bitset_test (set1, 511))
    abort ();
  if (!bitset_test (set1, 4095))
    abort ();

  bitset_set (set2, 2);
  bitset_set (set2, 513);
  bitset_set (set2, 4095);

  if (!bitset_test (set2, 2))
    abort ();
  if (!bitset_test (set2, 513))
    abort ();
  if (!bitset_test (set2, 4095))
    abort ();
}


static void
test_not (bitset set1, bitset set2, bitset set3)
{
  /* This test will make a sparse bitset full.  */
  bitset_zero (set1);
  bitset_not (set3, set1);
  bitset_not (set3, set3);
  if (!bitset_empty_p (set3))
    abort ();

  bitset_ones (set3);
  bitset_not (set3, set3);
  if (!bitset_empty_p (set3))
    abort ();
}


static void
test_or (bitset set1, bitset set2, bitset set3)
{
  test_set (set1, set2);

  bitset_or (set3, set1, set2);
    
  if (!bitset_subset_p (set3, set1))
    abort ();
  if (!bitset_subset_p (set3, set2))
    abort ();
  
  if (bitset_disjoint_p (set3, set1))
    abort ();  
  if (bitset_disjoint_p (set3, set2))
    abort ();  
}


static void
test_xor (bitset set1, bitset set2, bitset set3)
{
  test_set (set1, set2);

  bitset_xor (set3, set1, set1);
  if (!bitset_empty_p (set3))
    abort ();

  bitset_xor (set3, set2, set2);
  if (!bitset_empty_p (set3))
    abort ();
}


static void
test_and (bitset set1, bitset set2, bitset set3)
{
  test_set (set1, set2);

  bitset_and (set3, set1, set1);
  if (!bitset_equal_p (set3, set1))
    abort ();
    
  bitset_and (set3, set2, set2);
  if (!bitset_equal_p (set3, set2))
    abort ();

  bitset_and (set3, set1, set2);
  if (!bitset_only_set_p (set3, 4095))
    abort ();
}


static void
test_andn (bitset set1, bitset set2, bitset set3)
{
  test_set (set1, set2);

  bitset_andn (set3, set1, set1);
  if (!bitset_empty_p (set3))
    abort ();
    
  bitset_andn (set3, set2, set2);
  if (!bitset_empty_p (set3))
    abort ();

  bitset_zero (set1);
  bitset_andn (set3, set2, set1);
  if (!bitset_equal_p (set3, set2))
    abort ();
}


static void
test_and_or (bitset set1, bitset set2, bitset set3)
{
  test_set (set1, set2);

  bitset_zero (set3);
  bitset_and_or (set3, set3, set1, set2);
  if (!bitset_equal_p (set3, set2))
    abort ();

  bitset_zero (set3);
  bitset_and_or (set3, set1, set2, set3);
  bitset_and (set2, set2, set1);
    
  if (!bitset_equal_p (set3, set2))
    abort ();
}


static void
test_andn_or (bitset set1, bitset set2, bitset set3)
{
  test_set (set1, set2);

  bitset_zero (set3);
  bitset_andn_or (set3, set1, set3, set3);
  if (!bitset_equal_p (set3, set1))
    abort ();
}


static void
test_or_and (bitset set1, bitset set2, bitset set3)
{
  test_set (set1, set2);

  bitset_zero (set3);
  bitset_or_and (set3, set3, set1, set1);
  if (!bitset_equal_p (set3, set1))
    abort ();

  bitset_zero (set3);
  bitset_or_and (set3, set1, set2, set3);
  if (!bitset_empty_p (set3))
    abort ();
}


static void
test_misc (bitset set1, bitset set2, bitset set3)
{
  bitset_zero (set1);
  if (bitset_next (set1, 0) != -1)
    abort ();

  if (bitset_prev (set1, 0) != -1)
    abort ();

  bitset_set (set1, 127);
  if (bitset_next (set1, 0) != 127)
    abort ();
  if (bitset_prev (set1, 0) != 127)
    abort ();

  bitset_set (set1, 129);
  if (bitset_next (set1, 0) != 127)
    abort ();
  if (bitset_next (set1, 128) != 129)
    abort ();
  if (bitset_prev (set1, 0) != 129)
    abort ();
  if (bitset_next (set1, 3) != 127)
    abort ();
  if (bitset_prev (set1, 3) != 129)
    abort ();

  if (bitset_first (set1) != 127)
    abort ();
  if (bitset_last (set1) != 129)
    abort ();
}


static void
test (enum bitset_type type, unsigned int bits)
{
  bitset set1;
  bitset set2;
  bitset set3;
    
  set1 = bitset_alloc (bits, type);
  set2 = bitset_alloc (bits, type);
  set3 = bitset_alloc (bits, type);
    
  /* The sets should be created empty.  */
  if (!bitset_empty_p (set1))
    abort ();
  if (!bitset_empty_p (set2))
    abort ();
  if (!bitset_empty_p (set3))
    abort ();
    
  if (!bitset_disjoint_p (set1, set2))
    abort ();
    
  test_not (set1, set2, set3);
  
  test_or (set1, set2, set3);
  
  test_xor (set1, set2, set3);

  test_and (set1, set2, set3);

  test_andn (set1, set2, set3);

  test_or_and (set1, set2, set3);

  test_and_or (set1, set2, set3);

  test_andn_or (set1, set2, set3);

  test_misc (set1, set2, set3);

  bitset_free (set1);
  bitset_free (set2);
  bitset_free (set3);
}


int
main (void)
{
  bitset_stats_enable ();

  test (BITSET_ARRAY, 4096);
  test (BITSET_LIST, 4096);
  test (BITSET_TABLE, 4096);
  test (BITSET_VARRAY, 4096);

  test (BITSET_ARRAY, 4100);
  test (BITSET_LIST, 4100);
  test (BITSET_TABLE, 4100);
  test (BITSET_VARRAY, 4100);

  bitset_stats_dump (stderr);
  return 0;
}
