#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <bitset.h>


static double
random_uniform (double range, double mean)
{
    /* This returns a white uniformly distributed random number
       with a specified range and mean.  */

    return (drand48 () - 0.5) * range + mean;
}


static double 
random_gaussian (double variance, double mean)
{
    /* This returns a white Gaussian distributed random number with a
       specified variance and mean.  */

    double uniform1;
    double uniform2;
    
    uniform1 = drand48 ();  
    uniform2 = drand48 ();  
    return sqrt (-2.0 * variance * log (uniform1)) 
      * cos (2.0 * M_PI * uniform2) + mean;
}


static void
choose (unsigned int *seq, unsigned int bits, unsigned int rlsd)
{
  unsigned int i;
  bitset set;

  set = bitset_alloc (bits, BITSET_ARRAY);
  bitset_ones (set);

  i = 0;
  while (!bitset_empty_p (set))
    {
      int pos;
      int len;
      int start;
      int stop;
      int j;
      
      pos = random_uniform (bits, bits / 2);

      /* Choose run length.  */
      len = abs (random_gaussian (rlsd * rlsd, 0));

      start = pos - len / 2;
      stop = start + len - 1;
      if (start < 0)
	start = 0;
      if (stop >= bits)
	stop = bits - 1;

      len = stop - start + 1;
      if (len > (bits - i))
	len = bits - i;

      for (j = 0; j < len; j++)
	{
	  if (bitset_test (set, start + j))
	    {
	      bitset_reset (set, start + j);
	      seq[i++] = start + j;
	    }
	}
    }
}


static double
run (bitset set, unsigned int *seq, unsigned int bits)
{
  clock_t start;
  clock_t stop;
  int i;

  start = clock ();
  
  i = 0;
  while (!bitset_empty_p (set))
    {
      bitset_reset (set, seq[i++]);
      if (i > bits)
	abort ();
    }

  stop = clock ();

  if (i != bits)
    abort ();

  return (stop - start) / (double) CLOCKS_PER_SEC * 1000;
}


static void
test (enum bitset_type type, unsigned int bits, unsigned int *seq)
{
  const char *name;
  double runtime;
  bitset set1;
  int i;
  int trials = 10;

  set1 = bitset_alloc (bits, type);
  name = bitset_type_name_get (set1);

  runtime = 0;

  for (i = 0; i < trials; i++)
    {
      bitset_ones (set1);
      runtime += run (set1, seq, bits) / trials;
    }

  bitset_free (set1);

  printf ("%s: time = %.1f ms, bits = %d\n",
	  name, runtime, bits);
}


int
main (int argc, char **argv)
{
  unsigned int bits;
  unsigned int seed;
  unsigned int rlsd;
  unsigned int *seq;
  int do_all = 1;

  bits = argc > 1 ? atoi (argv[1]) : 1000000;
  sd = argc > 2 ? atoi (argv[2]) : 100;
  seed = argc > 3 ? atoi (argv[3]) : 0;

  srand48 (seed);
  
  // bitset_stats_enable ();

  seq = calloc (bits, sizeof (*seq));
  choose (seq, bits, rlsd);

  if (argc > 4)
    {
      do_all = 0;

      switch (argv[4][0])
	{
	case 'a':
	  test (BITSET_ARRAY, bits, seq);
	  break;
	case 'l':
	  test (BITSET_LIST, bits, seq);
	  break;
	case 'e':
	case 't':
	  test (BITSET_TABLE, bits, seq);
	  break;
	case 'v':
	  test (BITSET_VARRAY, bits, seq);
	  break;

	default:
	  do_all = 1;
	  break;
	}
    }

  if (do_all)
    {
      test (BITSET_ARRAY, bits, seq);
      test (BITSET_LIST, bits, seq);
      test (BITSET_TABLE, bits, seq);
      test (BITSET_VARRAY, bits, seq);
    }

  free (seq);

  bitset_stats_dump (stderr);
  return 0;
}
